@@ -23,10 +23,32 @@
 #include <rest/rest-proxy.h>
 #include <libsoup/soup.h>
 #include "oauth-proxy.h"
-#include "oauth-proxy-private.h"
 #include "oauth-proxy-call.h"
 
-G_DEFINE_TYPE (OAuthProxy, oauth_proxy, REST_TYPE_PROXY)
+typedef struct {
+  /* Application "consumer" keys */
+  char *consumer_key;
+  char *consumer_secret;
+  /* Authorisation "user" tokens */
+  char *token;
+  char *token_secret;
+  /* How we're signing */
+  OAuthSignatureMethod method;
+  /* OAuth 1.0a */
+  gboolean oauth_10a;
+  char *verifier;
+  /* OAuth Echo */
+  gboolean oauth_echo;
+  char *service_url;
+  /* URL to use for signatures */
+  char *signature_host;
+} OAuthProxyPrivate;
+
+struct _OAuthProxy {
+  RestProxy parent_instance;
+};
+
+G_DEFINE_TYPE_WITH_PRIVATE (OAuthProxy, oauth_proxy, REST_TYPE_PROXY)
 
 enum {
   PROP_0,
@@ -54,7 +76,8 @@ static void
 oauth_proxy_get_property (GObject *object, guint property_id,
                               GValue *value, GParamSpec *pspec)
 {
-  OAuthProxyPrivate *priv = PROXY_GET_PRIVATE (object);
+  OAuthProxy *self = OAUTH_PROXY (object);
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (self);
 
   switch (property_id) {
   case PROP_CONSUMER_KEY:
@@ -84,7 +107,8 @@ static void
 oauth_proxy_set_property (GObject *object, guint property_id,
                               const GValue *value, GParamSpec *pspec)
 {
-  OAuthProxyPrivate *priv = PROXY_GET_PRIVATE (object);
+  OAuthProxy *self = OAUTH_PROXY (object);
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (self);
 
   switch (property_id) {
   case PROP_CONSUMER_KEY:
@@ -123,7 +147,8 @@ oauth_proxy_set_property (GObject *object, guint property_id,
 static void
 oauth_proxy_finalize (GObject *object)
 {
-  OAuthProxyPrivate *priv = PROXY_GET_PRIVATE (object);
+  OAuthProxy *self = OAUTH_PROXY (object);
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (self);
 
   g_free (priv->consumer_key);
   g_free (priv->consumer_secret);
@@ -146,8 +171,6 @@ oauth_proxy_class_init (OAuthProxyClass *klass)
   RestProxyClass *proxy_class = REST_PROXY_CLASS (klass);
   GParamSpec *pspec;
 
-  g_type_class_add_private (klass, sizeof (OAuthProxyPrivate));
-
   object_class->get_property = oauth_proxy_get_property;
   object_class->set_property = oauth_proxy_set_property;
   object_class->finalize = oauth_proxy_finalize;
@@ -201,7 +224,8 @@ oauth_proxy_class_init (OAuthProxyClass *klass)
 static void
 oauth_proxy_init (OAuthProxy *self)
 {
-  PROXY_GET_PRIVATE (self)->method = HMAC_SHA1;
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (self);
+  priv->method = HMAC_SHA1;
 }
 
 /**
@@ -263,11 +287,11 @@ oauth_proxy_new (const char *consumer_key,
  */
 RestProxy *
 oauth_proxy_new_with_token (const char *consumer_key,
-                 const char *consumer_secret,
-                 const char *token,
-                 const char *token_secret,
-                 const gchar *url_format,
-                 gboolean binding_required)
+                            const char *consumer_secret,
+                            const char *token,
+                            const char *token_secret,
+                            const gchar *url_format,
+                            gboolean binding_required)
 {
   return g_object_new (OAUTH_TYPE_PROXY,
                        "consumer-key", consumer_key,
@@ -279,115 +303,6 @@ oauth_proxy_new_with_token (const char *consumer_key,
                        NULL);
 }
 
-typedef struct {
-  OAuthProxyAuthCallback callback;
-  gpointer user_data;
-} AuthData;
-
-static void
-auth_callback (RestProxyCall *call,
-               const GError  *error,
-               GObject       *weak_object,
-               gpointer       user_data)
-{
-  AuthData *data = user_data;
-  OAuthProxy *proxy = NULL;
-  OAuthProxyPrivate *priv;
-  GHashTable *form;
-
-  g_object_get (call, "proxy", &proxy, NULL);
-  priv = PROXY_GET_PRIVATE (proxy);
-
-  if (!error) {
-    /* TODO: sanity check response */
-    form = soup_form_decode (rest_proxy_call_get_payload (call));
-    priv->token = g_strdup (g_hash_table_lookup (form, "oauth_token"));
-    priv->token_secret = g_strdup (g_hash_table_lookup (form, "oauth_token_secret"));
-    g_hash_table_destroy (form);
-  }
-
-  data->callback (proxy, error, weak_object, data->user_data);
-
-  g_slice_free (AuthData, data);
-  g_object_unref (call);
-  g_object_unref (proxy);
-}
-
-/**
- * oauth_proxy_auth_step_async:
- * @proxy: an #OAuthProxy
- * @function: the function to invoke on the proxy
- * @callback: (scope async): the callback to invoke when authorisation is complete
- * @weak_object: the #GObject to weakly reference and tie the lifecycle too
- * @user_data: data to pass to @callback
- * @error_out: a #GError, or %NULL
- *
- * Perform an OAuth authorisation step.  This calls @function and then updates
- * the token and token secret in the proxy.
- *
- * @proxy must not require binding, the function will be invoked using
- * rest_proxy_call_set_function().
- */
-gboolean
-oauth_proxy_auth_step_async (OAuthProxy *proxy,
-                             const char *function,
-                             OAuthProxyAuthCallback callback,
-                             GObject *weak_object,
-                             gpointer user_data,
-                             GError **error_out)
-{
-  RestProxyCall *call;
-  AuthData *data;
-
-  call = rest_proxy_new_call (REST_PROXY (proxy));
-  rest_proxy_call_set_function (call, function);
-
-  data = g_slice_new0 (AuthData);
-  data->callback = callback;
-  data->user_data = user_data;
-
-  return rest_proxy_call_async (call, auth_callback, weak_object, data, error_out);
-  /* TODO: if call_async fails, the call is leaked */
-}
-
-/**
- * oauth_proxy_auth_step:
- * @proxy: an #OAuthProxy
- * @function: the function to invoke on the proxy
- * @error: return location for a #GError
- *
- * Perform an OAuth authorisation step.  This calls @function and then updates
- * the token and token secret in the proxy.
- *
- * @proxy must not require binding, the function will be invoked using
- * rest_proxy_call_set_function().
- */
-gboolean
-oauth_proxy_auth_step (OAuthProxy *proxy, const char *function, GError **error)
-{
-  OAuthProxyPrivate *priv = PROXY_GET_PRIVATE (proxy);
-  RestProxyCall *call;
-  GHashTable *form;
-
-  call = rest_proxy_new_call (REST_PROXY (proxy));
-  rest_proxy_call_set_function (call, function);
-
-  if (!rest_proxy_call_run (call, NULL, error)) {
-    g_object_unref (call);
-    return FALSE;
-  }
-
-  /* TODO: sanity check response */
-  form = soup_form_decode (rest_proxy_call_get_payload (call));
-  priv->token = g_strdup (g_hash_table_lookup (form, "oauth_token"));
-  priv->token_secret = g_strdup (g_hash_table_lookup (form, "oauth_token_secret"));
-  g_hash_table_destroy (form);
-
-  g_object_unref (call);
-
-  return TRUE;
-}
-
 /**
  * oauth_proxy_request_token:
  * @proxy: an #OAuthProxy
@@ -420,10 +335,11 @@ oauth_proxy_request_token (OAuthProxy *proxy,
   if (callback_uri)
     rest_proxy_call_add_param (call, "oauth_callback", callback_uri);
 
-  if (!rest_proxy_call_run (call, NULL, error)) {
-    g_object_unref (call);
-    return FALSE;
-  }
+  if (!rest_proxy_call_sync (call, error))
+    {
+      g_object_unref (call);
+      return FALSE;
+    }
 
   /* TODO: sanity check response */
   oauth_proxy_call_parse_token_response (OAUTH_PROXY_CALL (call));
@@ -434,37 +350,34 @@ oauth_proxy_request_token (OAuthProxy *proxy,
 }
 
 static void
-request_token_cb (RestProxyCall *call,
-                  const GError  *error,
-                  GObject       *weak_object,
-                  gpointer       user_data)
+request_token_cb (GObject      *source_object,
+                  GAsyncResult *result,
+                  gpointer      user_data)
 {
-  AuthData *data = user_data;
-  OAuthProxy *proxy = NULL;
+  GTask *task = G_TASK (user_data);
+  RestProxyCall *call = REST_PROXY_CALL (source_object);
+  GError *error = NULL;
+  gboolean call_status;
 
-  g_object_get (call, "proxy", &proxy, NULL);
-  g_assert (proxy);
+  call_status = rest_proxy_call_invoke_finish (call, result, &error);
 
-  if (!error) {
+  if (error != NULL) {
+    g_task_return_error (task, error);
+  } else {
     oauth_proxy_call_parse_token_response (OAUTH_PROXY_CALL (call));
+    g_task_return_boolean (task, call_status);
   }
 
-  data->callback (proxy, error, weak_object, data->user_data);
-
-  g_slice_free (AuthData, data);
-  g_object_unref (call);
-  g_object_unref (proxy);
+  g_object_unref (task);
 }
 
 /**
  * oauth_proxy_request_token_async:
  * @proxy: an #OAuthProxy
- * @function: the function name to invoke
- * @callback_uri: the callback URI
+ * @function: (nullable): the function name to invoke
+ * @callback_uri: (nullable): the callback URI
  * @callback: (scope async): a #OAuthProxyAuthCallback to invoke on completion
- * @weak_object: #GObject to weakly reference and tie the lifecycle of the method call too
  * @user_data: user data to pass to @callback
- * @error: a #GError, or %NULL
  *
  * Perform the Request Token phase of OAuth, invoking @function (defaulting to
  * "request_token" if @function is NULL).
@@ -475,21 +388,17 @@ request_token_cb (RestProxyCall *call,
  *
  * This method will return once the method has been queued, @callback will be
  * invoked when it has completed.
- *
- * Returns: %TRUE if the method was successfully queued, or %FALSE on
- * failure. On failure @error is set.
  */
-gboolean
-oauth_proxy_request_token_async (OAuthProxy            *proxy,
-                                 const char            *function,
-                                 const char            *callback_uri,
-                                 OAuthProxyAuthCallback callback,
-                                 GObject               *weak_object,
-                                 gpointer               user_data,
-                                 GError               **error)
+void
+oauth_proxy_request_token_async (OAuthProxy          *proxy,
+                                 const char          *function,
+                                 const char          *callback_uri,
+                                 GCancellable        *cancellable,
+                                 GAsyncReadyCallback  callback,
+                                 gpointer             user_data)
 {
   RestProxyCall *call;
-  AuthData *data;
+  GTask *task;
 
   call = rest_proxy_new_call (REST_PROXY (proxy));
   rest_proxy_call_set_function (call, function ? function : "request_token");
@@ -498,11 +407,33 @@ oauth_proxy_request_token_async (OAuthProxy            *proxy,
   if (callback_uri)
     rest_proxy_call_add_param (call, "oauth_callback", callback_uri);
 
-  data = g_slice_new0 (AuthData);
-  data->callback = callback;
-  data->user_data = user_data;
+  task = g_task_new (proxy, cancellable, callback, user_data);
 
-  return rest_proxy_call_async (call, request_token_cb, weak_object, data, error);
+  rest_proxy_call_invoke_async (call, cancellable, request_token_cb, task);
+
+  g_object_unref (call);
+}
+
+/**
+ * oauth_proxy_request_token_finish:
+ * @proxy: a #OAuthProxy
+ * @result: a #GAsyncResult
+ * @error: a #GError or %NULL
+ *
+ * Finishes an operation started with oauth_proxy_request_token_async()
+ *
+ * Returns: %TRUE on success, %FALSE if an error occurred, in which case
+ *   @error will be set.
+ */
+gboolean
+oauth_proxy_request_token_finish (OAuthProxy *proxy,
+                                  GAsyncResult *result,
+                                  GError **error)
+{
+  g_return_val_if_fail (OAUTH_IS_PROXY (proxy), FALSE);
+  g_return_val_if_fail (g_task_is_valid (result, proxy), FALSE);
+
+  return g_task_propagate_boolean (G_TASK (result), error);
 }
 
 /**
@@ -537,10 +468,11 @@ oauth_proxy_access_token (OAuthProxy *proxy,
   if (verifier)
     rest_proxy_call_add_param (call, "oauth_verifier", verifier);
 
-  if (!rest_proxy_call_run (call, NULL, error)) {
-    g_object_unref (call);
-    return FALSE;
-  }
+  if (!rest_proxy_call_sync (call, error))
+    {
+      g_object_unref (call);
+      return FALSE;
+    }
 
   /* TODO: sanity check response */
   oauth_proxy_call_parse_token_response (OAUTH_PROXY_CALL (call));
@@ -551,26 +483,25 @@ oauth_proxy_access_token (OAuthProxy *proxy,
 }
 
 static void
-access_token_cb (RestProxyCall *call,
-                 const GError  *error,
-                 GObject       *weak_object,
-                 gpointer       user_data)
+access_token_cb (GObject      *source_object,
+                 GAsyncResult *result,
+                 gpointer      user_data)
 {
-  AuthData *data = user_data;
-  OAuthProxy *proxy = NULL;
+  GTask *task = G_TASK (user_data);
+  RestProxyCall *call = REST_PROXY_CALL (source_object);
+  GError *error = NULL;
+  gboolean call_status;
 
-  g_object_get (call, "proxy", &proxy, NULL);
-  g_assert (proxy);
+  call_status = rest_proxy_call_invoke_finish (call, result, &error);
 
-  if (!error) {
+  if (error != NULL) {
+    g_task_return_error (task, error);
+  } else {
     oauth_proxy_call_parse_token_response (OAUTH_PROXY_CALL (call));
+    g_task_return_boolean (task, call_status);
   }
 
-  data->callback (proxy, error, weak_object, data->user_data);
-
-  g_slice_free (AuthData, data);
-  g_object_unref (call);
-  g_object_unref (proxy);
+  g_object_unref (task);
 }
 
 /**
@@ -579,9 +510,7 @@ access_token_cb (RestProxyCall *call,
  * @function: the function name to invoke
  * @verifier: the verifier
  * @callback: (scope async): a #OAuthProxyAuthCallback to invoke on completion
- * @weak_object: #GObject to weakly reference and tie the lifecycle of the method call too
  * @user_data: user data to pass to @callback
- * @error: a #GError, or %NULL
  *
  * Perform the Access Token phase of OAuth, invoking @function (defaulting to
  * "access_token" if @function is NULL).
@@ -593,21 +522,17 @@ access_token_cb (RestProxyCall *call,
  *
  * This method will return once the method has been queued, @callback will be
  * invoked when it has completed.
- *
- * Returns: %TRUE if the method was successfully queued, or %FALSE on
- * failure. On failure @error is set.
  */
-gboolean
-oauth_proxy_access_token_async (OAuthProxy            *proxy,
-                                const char            *function,
-                                const char            *verifier,
-                                OAuthProxyAuthCallback callback,
-                                GObject               *weak_object,
-                                gpointer               user_data,
-                                GError               **error)
+void
+oauth_proxy_access_token_async (OAuthProxy          *proxy,
+                                const char          *function,
+                                const char          *verifier,
+                                GCancellable        *cancellable,
+                                GAsyncReadyCallback  callback,
+                                gpointer             user_data)
 {
   RestProxyCall *call;
-  AuthData *data;
+  GTask *task;
 
   call = rest_proxy_new_call (REST_PROXY (proxy));
   rest_proxy_call_set_function (call, function ? function : "access_token");
@@ -616,11 +541,37 @@ oauth_proxy_access_token_async (OAuthProxy            *proxy,
   if (verifier)
     rest_proxy_call_add_param (call, "oauth_verifier", verifier);
 
-  data = g_slice_new0 (AuthData);
-  data->callback = callback;
-  data->user_data = user_data;
+  task = g_task_new (proxy, cancellable, callback, user_data);
 
-  return rest_proxy_call_async (call, access_token_cb, weak_object, data, error);
+  rest_proxy_call_invoke_async (call, cancellable, access_token_cb, task);
+  g_object_unref (call);
+}
+
+gboolean
+oauth_proxy_access_token_finish (OAuthProxy *proxy,
+                                 GAsyncResult *result,
+                                 GError **error)
+{
+  g_return_val_if_fail (OAUTH_IS_PROXY (proxy), FALSE);
+  g_return_val_if_fail (g_task_is_valid (result, proxy), FALSE);
+
+  return g_task_propagate_boolean (G_TASK (result), error);
+}
+
+const char *
+oauth_proxy_get_consumer_key (OAuthProxy *self)
+{
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (self);
+
+  return priv->consumer_key;
+}
+
+const char *
+oauth_proxy_get_consumer_secret (OAuthProxy *self)
+{
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (self);
+
+  return priv->consumer_secret;
 }
 
 /**
@@ -635,7 +586,8 @@ oauth_proxy_access_token_async (OAuthProxy            *proxy,
 const char *
 oauth_proxy_get_token (OAuthProxy *proxy)
 {
-  OAuthProxyPrivate *priv = PROXY_GET_PRIVATE (proxy);
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (proxy);
+
   return priv->token;
 }
 
@@ -649,14 +601,11 @@ oauth_proxy_get_token (OAuthProxy *proxy)
 void
 oauth_proxy_set_token (OAuthProxy *proxy, const char *token)
 {
-  OAuthProxyPrivate *priv;
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (proxy);
 
   g_return_if_fail (OAUTH_IS_PROXY (proxy));
-  priv = PROXY_GET_PRIVATE (proxy);
-
-  if (priv->token)
-    g_free (priv->token);
 
+  g_free (priv->token);
   priv->token = g_strdup (token);
 }
 
@@ -672,7 +621,8 @@ oauth_proxy_set_token (OAuthProxy *proxy, const char *token)
 const char *
 oauth_proxy_get_token_secret (OAuthProxy *proxy)
 {
-  OAuthProxyPrivate *priv = PROXY_GET_PRIVATE (proxy);
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (proxy);
+
   return priv->token_secret;
 }
 
@@ -686,10 +636,9 @@ oauth_proxy_get_token_secret (OAuthProxy *proxy)
 void
 oauth_proxy_set_token_secret (OAuthProxy *proxy, const char *token_secret)
 {
-  OAuthProxyPrivate *priv;
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (proxy);
 
   g_return_if_fail (OAUTH_IS_PROXY (proxy));
-  priv = PROXY_GET_PRIVATE (proxy);
 
   if (priv->token_secret)
     g_free (priv->token_secret);
@@ -710,9 +659,22 @@ oauth_proxy_set_token_secret (OAuthProxy *proxy, const char *token_secret)
 gboolean
 oauth_proxy_is_oauth10a (OAuthProxy *proxy)
 {
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (proxy);
+
   g_return_val_if_fail (OAUTH_IS_PROXY (proxy), FALSE);
 
-  return PROXY_GET_PRIVATE (proxy)->oauth_10a;
+  return priv->oauth_10a;
+}
+
+void
+oauth_proxy_set_oauth10a (OAuthProxy *self,
+                          gboolean    oauth10a)
+{
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (self);
+
+  g_return_if_fail (OAUTH_IS_PROXY (self));
+
+  priv->oauth_10a = oauth10a;
 }
 
 /**
@@ -727,10 +689,9 @@ oauth_proxy_is_oauth10a (OAuthProxy *proxy)
 const char *
 oauth_proxy_get_signature_host (OAuthProxy *proxy)
 {
-  OAuthProxyPrivate *priv;
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (proxy);
 
   g_return_val_if_fail (OAUTH_IS_PROXY (proxy), NULL);
-  priv = PROXY_GET_PRIVATE (proxy);
 
   return priv->signature_host;
 }
@@ -746,10 +707,9 @@ void
 oauth_proxy_set_signature_host (OAuthProxy *proxy,
                                 const char *signature_host)
 {
-  OAuthProxyPrivate *priv;
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (proxy);
 
   g_return_if_fail (OAUTH_IS_PROXY (proxy));
-  priv = PROXY_GET_PRIVATE (proxy);
 
   g_free (priv->signature_host);
 
@@ -777,14 +737,13 @@ oauth_proxy_new_echo_proxy (OAuthProxy  *proxy,
                             gboolean     binding_required)
 {
   OAuthProxy *echo_proxy;
-  OAuthProxyPrivate *priv, *echo_priv;
+  OAuthProxyPrivate *echo_priv;
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (proxy);
 
   g_return_val_if_fail (OAUTH_IS_PROXY (proxy), NULL);
   g_return_val_if_fail (service_url, NULL);
   g_return_val_if_fail (url_format, NULL);
 
-  priv = PROXY_GET_PRIVATE (proxy);
-
   echo_proxy = g_object_new (OAUTH_TYPE_PROXY,
                              "url-format", url_format,
                              "binding-required", binding_required,
@@ -794,7 +753,7 @@ oauth_proxy_new_echo_proxy (OAuthProxy  *proxy,
                              "token", priv->token,
                              "token-secret", priv->token_secret,
                              NULL);
-  echo_priv = PROXY_GET_PRIVATE (echo_proxy);
+  echo_priv = oauth_proxy_get_instance_private (echo_proxy);
 
   echo_priv->oauth_echo = TRUE;
   echo_priv->service_url = g_strdup (service_url);
@@ -819,3 +778,26 @@ oauth_signature_method_get_type (void)
   return enum_type_id;
 }
 
+gboolean
+oauth_proxy_is_echo (OAuthProxy *self)
+{
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (self);
+
+  return priv->oauth_echo;
+}
+
+const char *
+oauth_proxy_get_service_url (OAuthProxy *self)
+{
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (self);
+
+  return priv->service_url;
+}
+
+OAuthSignatureMethod
+oauth_proxy_get_sign_method (OAuthProxy *self)
+{
+  OAuthProxyPrivate *priv = oauth_proxy_get_instance_private (self);
+
+  return priv->method;
+}
